import json
import os
from itertools import islice
import numpy as np
import struct
import warnings

from nextnanopy.utils.datasets import Variable, Coord
from nextnanopy.utils.mycollections import DictList
from nextnanopy.utils.formatting import best_str_to_name_unit
from nextnanopy.utils.misc import get_filename, message_decorator, start_with_choice
from nextnanopy import defaults

import pyvista as pv


_msgs = defaults.messages['load_output']
load_message = lambda method: message_decorator(method, init_msg=_msgs[0], end_msg=_msgs[1])
def displayname(data):
    """
    Parameters
    ----------
    data: DataFolder or filepath
    Returns
    -------
    str
    formatting filenames and folder names in datafolder to display DataFolder in tree structure
    """
    if isinstance(data, DataFolder):
        return os.path.basename(data.fullpath)+r'/'
    else:
        return  os.path.basename(data)

class DataFolder(object):
    """
        This class stores information about output directory.
        The stored data contains files (.files) and folders (.folders)
        Navigation between folders could be done in 3 ways:
            1. DataFolder.folders['folder_name']
            2. DataFolder.go_to('subfolder1', 'subfolder2', 'subfolder3')
            3. DataFolder.subfolder1.subfolder2.subfolder3
        For each method see details below.

        The initialization of the class will execute load and create_navigation methods.


        Parameters:
        ----------------
        fullpath: str
            path to a file



        Attributes:
        ----------------
        fullpath: str
        path to a file

        folders: DictList
            subfolders of the DataFolder,
            keys: str
                names of subfolders
            values: DataFolder
                DataFolder objects of subfolders

        files: list
            paths to files in folder


        Methods:
        --------------
        load():
            load a DataFolder

        create_navigation:
            creates attributes for navigation like DataFolder.subfolder1.subfolder2.subfolder3
            if name of subfolder corresponds to existed attribute of DataFolder class, attribute will not be created!
            if name of subfolder constis spaces, dots or specials charecters, attribute will be created, but navigation
            to whis subfolder will not work  - attribute error.

        find(template, deep = False):
            searches for a files which names contain template.
            template shoud be string.
            if deep = True, searches in subfolders as well.

            return: list of files

        go_to(*args):
            goes to the location
            DataFolder_path\\arg1\\arg2\\arg3...

            if location is a file:
                return filepath
            if location is a folder:
                return DataFolder(location)

        filenames()
            return filenames of files in folder
        """
    def __init__(self, fullpath):
        if not os.path.isdir(fullpath):
            raise ValueError(f"{fullpath} is not a directory")
        self.fullpath = fullpath
        self.files = []
        self.folders = DictList()
        self.load()
        self.create_navigation()

    def load(self):
        list_of_nodes = os.listdir(self.fullpath)
        for node in list_of_nodes:
            node_path = os.path.join(self.fullpath, node)
            if os.path.isdir(node_path):
                new_folder = DataFolder(node_path)
                self.folders[node] = new_folder
            else:
                self.files.append(node_path)

    def create_navigation(self):
        check_list = dir(self)
        for key, folder in self.folders.items():
            if key in check_list:
                warnings.warn(f"foldername '{key}' is not availabel for attribute navigation."
                              " Please, use DataFolder.go_to()")
            else:
                setattr(self, key, folder)

    def find(self, template, deep=False):
        list_of_files = [file for file in self.files if template in os.path.basename(file)]
        if not deep:
            return list_of_files
        if not self.folders:
            return list_of_files
        else:
            for folder in self.folders:
                list_of_files += folder.find(template=template, deep=deep)
            return list_of_files

    def find_multiple(self, templates, deep=False):
        list_of_files = []

        for file in self.files:
            if all(keyword in os.path.basename(file) for keyword in templates):
                list_of_files.append(file)

        if deep and self.folders:
            for folder in self.folders:
                list_of_files += folder.find_multiple(templates=templates, deep=deep)

        return list_of_files



    def file(self, filename):
        matched_files = self.find(template = filename, deep = False)
        if not matched_files:
            raise ValueError(f'No file with filename {filename} in directory {self.fullpath}')
        elif len(matched_files) == 1:
            return matched_files[0]
        else:
            warnings.warn(f"More than one file match '{filename}' name match in directory {self.fullpath}. First match returned.")
            return matched_files[0]

    def go_to(self, *args):
        path = os.path.join(self.fullpath, *args)
        if os.path.isdir(path):
            data = DataFolder(path)
        elif os.path.isfile(path):
            data = path
        else:
            raise ValueError(f'No such file or directory {path}')
        return data

    def filenames(self):
        return [os.path.basename(file) for file in self.files]

    def __repr__(self):
        out = list()
        out.append(f'{self.__class__.__name__}')
        out.append(f'fullpath: {self.fullpath}')
        out.append(f'Folders: {len(self.folders)}')
        for key in self.folders.keys():
            out.append(key)
        out.append('Files:')
        out.append(str(self.filenames()))
        out = '\n'.join(out)
        return out


    def make_tree(self, level = 0, result = None, with_files = True, deep = True):
        if not result:
            result = []
        result.insert(level,level*'    '+ displayname(self))
        if with_files:
            for file in reversed(self.files):
                result.insert(level+1,(level+1)*'    '+displayname(file))
        level = level + 1
        for i in reversed(range(len(self.folders))):
            folder = self.folders[i]
            if deep:
                result = folder.make_tree(level, result = result, with_files = with_files)
            else:
                result.insert(level, level * '    ' + displayname(folder))
        return result


    def show_tree(self, with_files = True, deep = True):
        tree_list = self.make_tree(with_files = with_files, deep = deep)
        print('\n'.join(tree_list))

    @property
    def name(self):
        # folder_path = os.path.dirname(self.fullpath)
        folder_name = os.path.basename(self.fullpath)
        return folder_name

    def read_sweep_infodict(self):
        infodict_path = os.path.join(self.fullpath, 'sweep_infodict.json')

        if not os.path.isfile(infodict_path):
            raise FileNotFoundError(
                "Sweep infodict file not found. Make sure that the chosen folder is a sweep output folder.")

        with open(infodict_path, 'r') as file:
            infodict = json.load(file)

        return infodict







class Output(object):

    def __init__(self, fullpath, **loader_kwargs):
        self.fullpath = fullpath
        self.metadata = {}
        self.coords = DictList()
        self.variables = DictList()

    @property
    def folder(self):
        return os.path.split(self.fullpath)[0]

    @property
    def filename_only(self):
        return get_filename(self.fullpath, ext=False)

    @property
    def filename(self):
        return get_filename(self.fullpath, ext=True)

    @property
    def extension(self):
        return os.path.splitext(self.fullpath)[-1]

    @property
    def data(self):
        dl = DictList()
        dl.update(self.coords)
        dl.update(self.variables)
        return dl

    @load_message
    def load(self):
        pass

    def get_coord(self, key):
        return self.coords[key]

    def get_variable(self, key):
        return self.variables[key]

    def __getitem__(self, item):
        return self.data[item]

    def __setitem__(self, item, value):
        if item in self.coords.keys():
            self.coords[item] = value
        if item in self.variables.keys():
            self.variables[item] = value

    def __delitem__(self, item):
        if item in self.coords.keys():
            del self.coords[item]
        if item in self.variables.keys():
            del self.variables[item]

    def __repr__(self):
        out = []
        out.append(f'{self.__class__.__name__}')
        out.append(f'fullpath: {self.fullpath}')
        out.append(f'Coordinates: {len(self.coords)} datasets')
        for key, coord in self.coords.items():
            out.append(f'\t{str(coord)}')
        out.append(f'Variables: {len(self.variables)} datasets')
        for key, var in self.variables.items():
            out.append(f'\t{str(var)}')
        out = '\n'.join(out)
        return out

    def __iter__(self):
        self._iter_index = 0
        return self

    def __next__(self):
        try:
            result = self.data.__getitem__(self._iter_index)
        except (IndexError, KeyError):
            raise StopIteration
        self._iter_index += 1
        return result


class DataFileTemplate(Output):
    """
        This class stores the data from any kind of nextnano data files with the
        same structure.
        The stored data contains coordinates (.coords) and dependent variables (.variables).
        Each coordinate or variable would contain attributes like name, unit and value.
        For more information, see their specific documentation.

        The initialization of the class will execute the load method.

        ...

        Parameters
        ----------
        fullpath : str
            path to the file.


        Attributes
        ----------
        fullpath : str
            path to the file (default: None)
        coords : DictList
            Coord objects (default: DictList())
        variables : DictList
            Variable objects (default: DictList())
        data : DictList
            coords and variables together
        metadata : dict
            extra information
        filename : str
            name with the file extension
        filename_only :
            name without the file extension
        extension : str
            file extension
        folder : str
            folder of the fullpath
        product : str
            flag about nextnano product to help to find the best loading routine


        Methods
        -------
        load(fullpath)
            load a data file

        get_coord(name)
            equivalent to self.coords[name]

        get_variable(name)
            equivalent to self.variables[name]

    """

    def __init__(self, fullpath, product=None, **loader_kwargs):
        super().__init__(fullpath)
        self.product = product
        self.load(**loader_kwargs)

    @load_message
    def load(self, **loader_kwargs):
        """
        Find the loader and update the stored information with the loaded data
        """
        loader = self.get_loader()
        df = loader(self.fullpath, **loader_kwargs)
        self.update_with_datafile(df)
        del df

    def update_with_datafile(self, datafile):
        """
        Copy .metadata, .coords and .variables of the specified datafile
        Copy other attributes like .vtk if there is any.

        Parameters
        ----------
        datafile : nextnano.outputs.DataFileTemplate object
        """

        self.metadata = datafile.metadata
        self.coords = datafile.coords
        self.variables = datafile.variables
        if hasattr(datafile, 'vtk'):
            self.vtk = datafile.vtk

    def get_loader(self):
        pass

    def export(self, filename, format):
        raise NotImplementedError('Exporters are not implemented yet')


class DataFile(DataFileTemplate):
    def __init__(self, fullpath, product=None, **loader_kwargs):
        super().__init__(fullpath, product=product, **loader_kwargs) # **loader_kwargs) #, FirstVarIsCoordFlag = False

    def get_loader(self):
        if self.product:
            loader = defaults.get_DataFile(self.product)
        else:
            print('[Warning] nextnano product is not specified: nextnano++, nextnano3, nextnano.NEGF or nextnano.MSB')
            print('[Warning] Autosearching for the best loading method. Note: The result may not be correct')
            loader = self.find_loader()
        return loader

    def find_loader(self):
        from nextnanopy.nnp.outputs import DataFile as DataFile_nnp
        from nextnanopy.nn3.outputs import DataFile as DataFile_nn3
        from nextnanopy.negf.outputs import DataFile as DataFile_negf
        from nextnanopy.msb.outputs import DataFile as DataFile_msb

        Dats = [DataFile_nn3, DataFile_nnp, DataFile_negf, DataFile_msb]
        for Dati in Dats:
            try:
                df = Dati(self.fullpath)
                if '' in df.variables.keys():
                    continue
                else:
                    break
            except:
                pass
        loader = Dati
        return loader

    def plot(self, legend = False, y_axis_name = '', subplots = False):
        import matplotlib.pyplot as plt
        if len(self.coords) == 0:
            fig, ax = plt.subplots()
            if len(self.variables)>1:
                plot_coord =self.variables[0]
                #plot_var = self.variables[1:]
                #for var in plot_var:
                for i in range(1, len(self.variables)-1):
                    var = self.variables[i]
                    ax.plot(plot_coord.value, var.value, label=var.name)
                ax.set_xlabel(f'{plot_coord.name}[{plot_coord.unit}]')
            else:
                for var in self.variables:
                    ax.plot(var.value, label=var.name)
            ax.set_ylabel(f'{y_axis_name}[{var.unit}]')
        elif len(self.coords) == 1:
            fig, ax = plt.subplots()
            x_coord = self.coords[0]
            x_label_name = x_coord.name
            x_label_unit = x_coord.unit
            x_value = x_coord.value
            ax.set_xlabel(f'{x_label_name}[{x_label_unit}]')
            for var in self.variables:
                ax.plot(x_value, var.value, label = var.name)
            ax.set_ylabel(f'{y_axis_name}[{var.unit}]')
            if legend:
                plt.legend()
        elif len(self.coords) == 2:
            x = self.coords[0]
            y = self.coords[1]
            number_of_var = len(self.variables)
            if number_of_var <2:
                fig, ax = plt.subplots()
                ax.set_xlabel(f'{x.name}[{x.unit}]')
                ax.set_ylabel(f'{y.name}[{y.unit}]')
                var = self.variables[0]
                im = ax.pcolormesh(x.value, y.value, var.value, shading='auto', label=f'{var.name}[{var.unit}]')
                ax.set_title(f'{var.name}[{var.unit}]')
                fig.colorbar(im, ax = ax)
            else:
                if subplots:
                    fig, ax = plt.subplots(len(self.variables))
                    for i, var in enumerate(self.variables):
                        im = ax[i].pcolormesh(x.value, y.value, var.value, shading='auto', label=f'{var.name}[{var.unit}]')
                        ax[i].set_xlabel(f'{x.name}[{x.unit}]')
                        ax[i].set_ylabel(f'{y.name}[{y.unit}]')
                        ax[i].set_title(f'{var.name}[{var.unit}]')
                        fig.colorbar(im, ax=ax[i])
                else:
                    for i, var in enumerate(self.variables):
                        fig, ax = plt.subplots()
                        im = ax.pcolormesh(x.value, y.value, var.value, shading='auto', label=f'{var.name}[{var.unit}]')
                        ax.set_xlabel(f'{x.name}[{x.unit}]')
                        ax.set_ylabel(f'{y.name}[{y.unit}]')
                        ax.set_title(f'{var.name}[{var.unit}]')
                        fig.colorbar(im, ax=ax)
        else:
            raise NotImplementedError('Preview plot feature is implemented only for datafiles with 2 or less coordinates')
        return fig,ax

    def save(self, filepath, format='dat'):
        """
           Save the data from the DataFile instance to a specified file in various formats.

           Parameters:
               filepath (str): The file path where the data will be saved.
               format (str, optional): The format in which the data should be saved. Default is 'dat'.
                   Supported formats:
                   - 'dat' (for 1D files): Data is saved in a plain text format (.dat) with whitespace-separated values.
                   - 'VTKAscii' (for 2D/3D files): Data is saved in VTK ASCII format (.vtk) suitable for visualization tools.
                   - 'AvsAscii_one_file' (for 2D/3D files): Data is saved in AVS/Express ASCII format (.fld) for AVS/Express software.

           Raises:
               NotImplementedError: If the provided 'format' is not supported for saving.

           Notes:
               - For the 'dat' format, the DataFile instance should be one-dimensional (ndim=1).
               - The 'VTKAscii' format requires 'nextnanopy.utils.formatting' module for creating the VTK header.
               - The 'AvsAscii_one_file' format requires the 'write_avsascii_one_file' function.

           Example:
               # Assuming `data_file` is an instance of the `DataFile` class
               data_file.save('data_file.dat', format='dat')  # Save data in .dat format
               data_file.save('data_file.vtk', format='VTKAscii')  # Save data in VTK ASCII format
               data_file.save('data_file.fld', format='AvsAscii_one_file')  # Save data in AVS/Express ASCII format
           """
        # TODO  and AvsBinary (.fld)
        accepted_format = ['dat', 'VTKAscii', 'AvsAscii_one_file']
        if format not in accepted_format:
            raise NotImplementedError(f'{format} format is not supported for saving')
        if format=='dat':
            ndim = len(self.coords)
            if ndim>1:
                raise ValueError(f'The DataFile is {ndim}-dimensional. More than 1 dimension is not supported for .dat extension')
            with open(filepath, 'w') as f:
                # Write headers
                for header in self.metadata['headers']:
                    f.write(header)

                # Write data
                combined_data = [coord.value for coord in self.coords] + [variable.value for variable in self.variables]

                data = np.column_stack(combined_data).transpose()
                np.savetxt(f, data.T)
        elif format=='VTKAscii':
            from nextnanopy.utils.formatting import create_vtk_header
            header = create_vtk_header(len(self.coords), [len(coord.value)  for coord in self.coords])
            with open(filepath, 'w') as file:
                # Write the header
                file.write(header)

                # Write the coordinates
                file.write('<Coordinates>')
                for coord in self.coords:
                    file.write(
                        f'<DataArray type="Float64" Name="{coord.name.upper()}_COORDINATES" NumberOfComponents="1" format="ascii">\n'
                            )
                    np.savetxt(file, coord.value, fmt='%.6f', delimiter='\t')
                    file.write('</DataArray>\n')
                if len(self.coords)==2: #add Z coordinate
                    file.write("""<DataArray type="Float64" Name ="Z_COORDINATES" NumberOfComponents ="1" format="ascii">
                                        0
                                    </DataArray>\n""")
                file.write('</Coordinates>\n')

                # Write the data
                file.write('<PointData>\n')
                for variable in self.variables:
                    if variable.unit:
                        name = variable.name+f'[{variable.unit}]'
                    else:
                        name = variable.name
                    file.write(f'<DataArray type="Float64" Name="{name}" NumberOfComponents="1" format="ascii">\n')
                    flatten_array = variable.value.flatten('F')

                    np.savetxt(file, flatten_array, fmt='%.6f', delimiter='\t')
                    file.write('</DataArray>\n')

                # Write the footer
                footer = """</PointData>
                            </Piece>
                            </RectilinearGrid>
                            </VTKFile>"""
                file.write(footer)
        elif format=='AvsAscii_one_file':
            write_avsascii_one_file(coordinates=self.coords, variables=self.variables, filename=filepath)
        self.filepath = filepath




class AvsAscii(Output):
    def __init__(self, fullpath, **loader_kwargs):
        super().__init__(fullpath)
        self.load()

    @property
    def fld(self):
        filename = self.filename_only + '.fld'
        return os.path.join(self.folder, filename)

    def load(self):
        self.load_raw_metadata()
        self.load_metadata()
        self.load_variables()
        self.load_coords()

    def load_raw_metadata(self):
        possible_keys = ['ndim', 'dim', 'nspace', 'veclen','data', 'field', 'label', 'variable', 'coord']
        info = []
        try:
            with open(self.fld, 'r') as f:
                for line in f:
                    line = line.replace('\n', '')
                    line = line.strip()
                    try:
                        float(line)
                        break
                    except:
                        if line == '':
                            continue
                        # if line[0] != '#':
                        #     info.append(line)
                        if start_with_choice(line,*possible_keys):
                            info.append(line)
        except UnicodeDecodeError:
            with open(self.fld,'rb') as f:
                for line in f:
                    try:
                        line = line.decode('ascii')
                    except:
                        break
                    line = line.replace('\n', '')
                    line = line.strip()        
                    if line == '':
                        continue
                    # if line[0] != '#':
                    #     info.append(line)
                    if start_with_choice(line, *possible_keys):
                        info.append(line)
        return info

    def load_metadata(self):
        info = self.load_raw_metadata()
        key_int = ['ndim', 'dim1', 'dim2', 'dim3', 'nspace', 'veclen']
        key_str = ['data', 'field']
        metadata = {}
        metadata['labels'] = []
        metadata['units'] = []
        metadata['variables'] = []
        metadata['coords'] = []
        metadata['dims'] = []
        for line in info:
            key, value = line.split(maxsplit=1)
            if value[0] == '=':
                value = value.replace('=', '')
                value = value.strip()
                if key in key_int:
                    value = int(value)
                    metadata[key] = value
                elif key == 'label':
                    labels = value.split()
                    for label in labels:
                        label, unit = best_str_to_name_unit(label, default_unit=None)
                        metadata['labels'].append(label)
                        metadata['units'].append(unit)
                else:
                    value = str(value)
                    metadata[key] = value

                if key[:3] == 'dim':
                    metadata['dims'].append(metadata[key])

            else:
                if key == 'variable':
                    vm = values_metadata(line)
                    vm['file'] = os.path.join(self.folder, vm['file'])
                    vm['size'] = np.prod(metadata['dims'])
                    metadata['variables'].append(vm)
                elif key == 'coord':
                    vm = values_metadata(line)
                    vm['file'] = os.path.join(self.folder, vm['file'])
                    num = vm['num']
                    vm['size'] = metadata[f'dim{num}']
                    metadata['coords'].append(vm)

        self.metadata = metadata
        return metadata

    def load_variables(self):
        meta = self.metadata
        variables = DictList()
        for vmeta, label, unit in zip(meta['variables'], meta['labels'], meta['units']):
            values = load_values(file=vmeta['file'],
                                 filetype=vmeta['filetype'],
                                 datatype = meta['data'],
                                 skip=vmeta['skip'],
                                 offset=vmeta['offset'],
                                 stride=vmeta['stride'],
                                 size=vmeta['size'])
            values = reshape_values(values, *meta['dims'])
            var = Variable(name=label, value=values, unit=unit, metadata=vmeta)
            variables[var.name] = var
        self.variables = variables
        return variables

    def load_coords(self):
        meta = self.metadata
        coords = DictList()
        for vmeta in meta['coords']:
            values = load_values(file=vmeta['file'],
                                 filetype=vmeta['filetype'],
                                 datatype = meta['data'],
                                 skip=vmeta['skip'],
                                 offset=vmeta['offset'],
                                 stride=vmeta['stride'],
                                 size=vmeta['size'])
            ax = coord_axis(vmeta['num'])
            unit = 'nm'
            var = Coord(name=ax, value=values, unit=unit, dim=vmeta['num'] - 1, metadata=vmeta)
            coords[var.name] = var
        self.coords = coords
        return coords


class Vtk(Output):
    def __init__(self, fullpath, **loader_kwargs):
        super().__init__(fullpath)
        self.load()

    def load(self):
        self.vtk = pv.read(self.fullpath)
        self.load_coords()
        self.load_variables()

    def load_coords(self):
        for i, coord in enumerate(['x', 'y', 'z']):
            if not hasattr(self.vtk, coord):
                continue
            value = getattr(self.vtk, coord)
            if value.size == 1:
                continue
            self.coords[coord] = Coord(name=coord, value=value, unit=None, dim=i)

    def load_variables(self):
        for _name in self.vtk.array_names:
            name, unit = best_str_to_name_unit(_name, default_unit=None)
            value = np.array(self.vtk[_name]).reshape(self.vtk.dimensions, order='F').squeeze()
            self.variables[name] = Variable(name=name, value=value, unit=unit)


def coord_axis(dim):
    dim = str(dim)
    axes = {'1': 'x', '2': 'y', '3': 'z'}
    return axes[dim]


def values_metadata(line):
    """ Return a dict for: kind, num, file, filetype, skip, offset, stride"""
    metadata = {}
    kind, num, rest = line.split(maxsplit=2)
    metadata['kind'] = kind
    metadata['num'] = int(num)
    raw_rest = rest.split('=')
    raw_rest = [r.strip().split() for r in raw_rest]
    rest = []
    for ri in raw_rest:
        rest.extend(ri)
    keys = rest[0::2]
    values = rest[1::2]
    for key, value in zip(keys, values):
        key = key.strip()
        value = value.strip()
        if key in ['num', 'skip', 'offset', 'stride']:
            value = int(value)
        metadata[key] = value
    return metadata


def load_values(file, filetype='ascii',datatype = 'double', skip=0, offset=0, stride=1, size=None):
    """ Return flat array of floating values """
    if filetype == 'ascii':
        
        stop = skip + size if size != None else None
        with open(file, 'r') as f:
            lines = islice(f, skip, stop, 1)
            values = [line.replace('\n', '').strip().split()[offset] for line in lines]
    elif filetype == 'binary':
        datatypes = {'double':'d'}
        datatype_sizes = {'double':8}
        datatype_size = datatype_sizes[datatype]
        with open(file,'rb') as datafile:
            datafile.seek(skip)
            data = datafile.read(size*datatype_size)
            values = []
            iteration = struct.iter_unpack(datatypes[datatype],data)
            for i in iteration:
                values.append(i[0])
            
    else:
        raise ValueError('filetype is not recognized or implemented')
    return np.array(values, dtype=float)


def reshape_values(values, *dims):
    dims = np.flip(dims)
    shape = tuple([dim for dim in dims])
    values = np.reshape(values, shape)
    return np.transpose(values)

def write_avsascii_one_file(coordinates, variables, filename, binary=False):
    if binary:
        filetype = 'binary'
        numpy_encoding = 'bytes'
    else:
        filetype = 'ascii'
        numpy_encoding = 'bytes'
    # Validate input
    if len(coordinates) < 2 or len(coordinates) > 3:
        raise ValueError("Coordinates array must have exactly 2 or 3 objects.")

    num_coords = len(coordinates)
    num_variables = len(variables)

    # Calculate skip values
    # TODO calculate skip values (in bytes?) for binary output
    header_up_to_first_label = 9
    header_lines = header_up_to_first_label + num_variables + 1 + num_variables + num_coords + 1 # each 1 is skipped line
    coord_skip_values = [header_lines]
    for i in range(num_coords-1):
        coord_skip_values.append(coord_skip_values[-1] + coordinates[i].value.size+1) #1 for empty line after each coord

    data_skip = coord_skip_values[-1] + coordinates[-1].value.size+1

    # Calculate the number of lines each variable takes
    number_of_lines_for_var = np.prod([coord.value.size for coord in coordinates])

    # Open file for writing
    with open(filename, "w") as file:
        # Write header
        file.write("# AVS/Express field file\n")
        file.write("#\n")
        file.write(f"ndim = {num_coords}\n")

        for i in range(num_coords):
            file.write(f"dim{i+1} = {coordinates[i].value.size}\n")

        file.write("nspace = {}\n".format(num_coords))
        file.write(f"veclen = {num_variables}\n")
        file.write("data = double\n")
        file.write("field = rectilinear\n")

        # Write variable labels and units
        for i in range(num_variables):
            file.write(f"label = {variables[i].name}[{variables[i].unit}]\n")

        # Write an empty line after the last label
        file.write("\n")

        # Write variable file paths
        for i, var in enumerate(variables):
            file.write(f"variable {i + 1} file={filename} filetype={filetype} skip={data_skip + i * number_of_lines_for_var} offset=0 stride=1\n")

        # Write coordinate file paths
        for coord, skip_value, i in zip(coordinates, coord_skip_values, range(num_coords)):
            file.write(f"coord {i+1} file={filename} filetype={filetype} skip={skip_value} offset=0 stride=1\n")

        # Write an empty line after the last coord line
        file.write("\n")

        # Write data for coordinates
        for coord in coordinates:
            np.savetxt(file, coord.value.flatten('F'), fmt='%.8f', encoding=numpy_encoding)
            file.write("\n")

        # Write data for variables
        for var in variables:
            np.savetxt(file, var.value.flatten('F'), fmt='%.8f', encoding=numpy_encoding)



import os
import numpy as np

def is_file(fullpath):
    filename = os.path.split(fullpath)[-1]
    if '.' not in filename:
        bool = False
    else:
        bool = True
    return bool


def get_filename(fullpath, ext=True):
    if not is_file(fullpath):
        raise ValueError(f'{fullpath} is not a file')
    filename = os.path.split(fullpath)[-1]
    if not ext:
        filename = os.path.splitext(filename)[0]
    return filename


def get_file_extension(fullpath):
    if not is_file(fullpath):
        raise ValueError(f'{fullpath} is not a file')
    filename = os.path.split(fullpath)[-1]
    ext = f".{filename.split('.')[-1]}"
    return ext


def get_folder(fullpath):
    if not is_file(fullpath):
        folder = fullpath
    else:
        folder = os.path.split(fullpath)[0]
    return folder


def get_path_files(path):
    if path == '':
        path = '.'
    files = [f for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))]
    return files


def mkdir_if_not_exist(path):
    from pathlib import Path
    Path(path).mkdir(parents=True, exist_ok=True)
    return path


def find_unused_name(name, list_names, extension, max_idx=True):
    if extension[0] != '.':
        extension = f'.{extension}'
    if extension not in name:
        name += extension
    prefix = get_file_prefix(name)
    lnames = list(filter(lambda ln: extension in ln and prefix in ln, list_names))

    if len(lnames) == 0:
        max_idx = 0
    if max_idx:
        idxs = [get_file_idx(name) for name in lnames]
        idxs.sort()
        max_idx = idxs[-1]
        unused_name = f'{prefix}_{max_idx + 1}{extension}'
    else:
        idx = 0
        unused_name = f'{prefix}_{idx}{extension}'
        while unused_name in list_names:
            unused_name = f'{prefix}_{idx}{extension}'
            idx += 1
    return unused_name


def find_unused_in_folder(fullpath, overwrite=False):
    #folder, name = os.path.split(fullpath)
    name = os.path.basename(fullpath)
    folder = os.path.dirname(fullpath)
    ext = get_file_extension(name)
    cwd_files = get_path_files(folder)
    if not overwrite:
        name = find_unused_name(name, cwd_files, ext)
    return os.path.join(folder, name)


def savetxt(fullpath, text, overwrite=False, automkdir=True):
    if automkdir:
        mkdir_if_not_exist(get_folder(fullpath))
    fullpath = find_unused_in_folder(fullpath, overwrite)
    with open(fullpath, 'w+') as file:
        file.write(text)
    return fullpath


def get_file_prefix(file):
    prefix, ext = os.path.splitext(file)
    idx = get_file_idx(prefix)
    if idx > -1:
        prefix = '_'.join(prefix.split('_')[0:-1])
    return prefix


def get_file_idx(file):
    prefix, ext = os.path.splitext(file)
    idx = -1
    if '_' in prefix:
        end = prefix.split('_')[-1]
        try:
            idx = int(end)
        except:
            pass
    return idx


def message_decorator(method, init_msg=None, end_msg=None):
    def f(*args, **kwargs):
        show_message(init_msg)
        result = method(*args, **kwargs)
        show_message(end_msg)
        return result

    return f


def show_message(msg):
    if msg is None:
        return
    if type(msg) == str:
        print(msg)
    elif callable(msg):
        msg()
    else:
        pass

def mkdir_even_if_exists(path,name):
    """creates a directory under path with a given name. If exists, adds integer number to directory name.
    returns directory full path"""
    directory = os.path.join(path,name)
    if os.path.exists(directory):
        i = 0
        while True:
            directory_numbered = directory +str(i)
            if os.path.exists(directory_numbered):
                i+=1
            elif i>2147483646:
                raise StopIteration('too many folders with the same name')
            else:
                os.makedirs(directory_numbered)
                directory = directory_numbered
                break
    else:
        os.makedirs(directory)
    return directory


def combinations(*args):
    n = len(args)
    array_of_combinations = np.array(np.meshgrid(*args)).T.reshape(-1,n)
    return array_of_combinations

def start_with_choice(line : str, *args : str):
    for arg in args:
        if line.startswith(args):
            return True
    return False


import sys, os
import subprocess
import queue
import threading
import warnings
from nextnanopy.utils.misc import get_filename, mkdir_if_not_exist
from nextnanopy import defaults

from nextnanopy.utils.formatting import generate_command

def command(
        inputfile,
        exe,
        license,
        database,
        outputdirectory,
        **opt_kwargs,
):
    kwargs = dict(
        inputfile=inputfile,
        exe=exe,
        license=license,
        database=database,
        outputdirectory=outputdirectory,
    )
    kwargs.update(opt_kwargs)
    product = defaults.input_file_type(inputfile)
    cmd = defaults.get_command(product)

    # warn if output path might be too long
    tooLongPath     = (product in ['nextnano3', 'nextnano++']) and len(outputdirectory) + 80 > 260   # TODO: how long is the minimal path appended by nn3/nnp simulations?
    tooLongPathNEGF = (product in ['nextnano.NEGF', 'nextnano.NEGF++']) and len(outputdirectory) + 80 > 260
    if tooLongPath or tooLongPathNEGF:
        warnings.warn('The output path might be too long on Windows 10 (maximum 260 characters). Consider abbreviating your input file name and/or sweep variables...')
    return cmd(**kwargs)


def send(cmd, cwd = os.getcwd()):
    PIPE = subprocess.PIPE
    return subprocess.Popen(cmd, stdout=PIPE, stderr=PIPE, close_fds=True, shell=True, cwd = cwd)


def read_output(pipe, funcs):
    for line in iter(pipe.readline, b''):
        for func in funcs:
            func(line)
    pipe.close()


def write_output(get, filepath, show=True):
    with open(filepath, 'w', newline='') as f:
        for line in iter(get, None):
            line = str(line, 'utf-8', errors='ignore')
            if show:
                sys.stdout.write(line)
            f.write(line)



def start_log(process, filepath, show=True, parallel =False):
    q = queue.Queue()
    out, err = [], []
    tout = threading.Thread(
        target=read_output, args=(process.stdout, [q.put, out.append]))
    terr = threading.Thread(
        target=read_output, args=(process.stderr, [q.put, err.append]))
    twrite = threading.Thread(target=write_output, args=(q.get, filepath, show))
    for t in (tout, terr, twrite):
        t.daemon = False
        t.start()


    if parallel:
        return q, tout, terr
    else:
        process.wait()
        for t in (tout, terr):
            t.join()
        q.put(None)
        return q, tout, terr



def execute(
        inputfile,
        exe,
        license,
        database,
        outputdirectory,
        show_log=True,
        parallel = False,
        **kwargs,
):
    filename = get_filename(inputfile, ext=False)
    inputfile = os.path.abspath(inputfile)
    outputdirectory = os.path.join(outputdirectory, filename)
    mkdir_if_not_exist(outputdirectory)
    logfile = os.path.join(outputdirectory, f'{filename}.log')
    cmd = command(inputfile, exe, license, database, outputdirectory, **kwargs)
    cwd = os.getcwd()
    wdir, executable = os.path.split(exe)  # nn3 assumes wdir at one folder upper than the executable
    
    # validate configuration of executable path
    if executable == '':
        raise FileNotFoundError(f'Executable path is empty! Check nextnanopy.config')


    if (not os.path.isfile(exe)) or (not os.path.isdir(wdir)):
        raise FileNotFoundError(f'Executable path is invalid: {exe}\nCheck nextnanopy.config')
    process = send(cmd, cwd = wdir)
    queue, tout, terr = start_log(process, logfile, show_log, parallel = parallel)
    os.chdir(cwd)
    info = {
        'process': process,
        'outputdirectory': outputdirectory,
        'filename': filename,
        'logfile': logfile,
        'cmd': cmd,
        'wdir': wdir,
        'queue': queue,
        'tout': tout,
        'terr': terr,
    }
    return info

def run_script(script, kwargs = None, show_log = True):
    """
    The function runs a python script with given arguments. Output is stored in the file script_name.log
    Parameters
    ----------
    script: str
        path to the python script
    kwargs: dict
        optional parameters, {keyword:argument,keyword:argument}
        for a keyword without argument leave argument as an empty string ''

        EXAMPLE: {'-o': 'my_output_folder','-p':''} will be converted to '-o my_output_folder -p'
    show_log - bool
        show the log in console output, default is True

    Returns
    -------
    process: subprocess.POPEN
    """
    args = [[sys.executable, script]]
    if kwargs:
        for key in kwargs.keys():
            args.append([key,kwargs[key]])
    cmd = generate_command(args)
    process = send(cmd)
    logfile = os.path.join(os.getcwd(),f'{os.path.basename(script)}.log')
    start_log(process, logfile, show_log)
    return process


import os,sys
import threading
import time
import warnings
import itertools
import queue
import nextnanopy
from nextnanopy.utils.formatting import text_to_lines, lines_to_text
from nextnanopy.utils.mycollections import DictList
from nextnanopy.utils.misc import savetxt, get_filename, get_folder, get_file_extension, message_decorator, mkdir_even_if_exists, mkdir_if_not_exist
from nextnanopy.commands import execute as cmd_execute
from nextnanopy import defaults
from collections.abc import Iterable


_msgs = defaults.messages['load_input']
load_message = lambda method: message_decorator(method, init_msg=_msgs[0], end_msg=_msgs[1])
_msgs = defaults.messages['save_input']
save_message = lambda method: message_decorator(method, init_msg=_msgs[0], end_msg=_msgs[1])
_msgs = defaults.messages['execute_input']
execute_message = lambda method: message_decorator(method, init_msg=_msgs[0], end_msg=_msgs[1])


class InputFileTemplate(object):
    """
        This class stores and manipulates any kind of nextnano input files.
        For each nextnano product, the syntax is different but the core information
        is very similar like the variables that can be changed.
        This class contains useful methods such as to get a preview of the file or
        execute a simulation.

        The initialization of the class will execute the load method and setup the
        config file.

        ...

        Parameters
        ----------
        fullpath : str
            path to the file.
            If it is not None, it will load the file (default: None)
        configpath : str
            path to the config file.
            If it is None, it will use the default configuration (default: None)


        Attributes
        ----------

        fullpath : str
            path to the file (default: None)
        configpath : str
            path to the config file (default: your home path)
        variables : DictList
            input variables defined in the file (default: DictList())
        raw_lines : list
            list of str of each line in the file
        text : str
            raw text of the file (settable)
        lines : list
            raw_lines updated with the current input variable parameters
        filename : str
            name with the file extension (settable)
        filename_only :
            name without the file extension (settable)
        folder_input : str
            folder of the fullpath (settable)
        folder_output : str
            folder where the simulated data is stored after execution
        product : str
            detected nextnano product when the file is loaded (default: 'not valid')
        config: nextnano.NNConfig
            config file object
        execute_info: dict
            information after executing the file

        Methods
        -------
        preview(nums=True)
            print the text of the file.

        load(fullpath)
            load an input file

        save(fullpath=None, overwrite=False, automkdir=True)
            save the current information into a file.

        execute(**kwargs)
            execute the input file located at .fullpath

        get_variable(name)
            equivalent to self.variables[name]

        set_variable(name, value=None, comment=None)
            change the value and/or the comment of self.variable[name]
            If value or comment is None, it won't change that parameter
    """

    def __init__(self, fullpath=None, configpath=None):
        self.raw_lines = []
        self.variables = DictList()
        self.content = None
        self.fullpath = fullpath
        self.product = 'not valid'
        self.__parallel__ = False
        if fullpath is not None:
            self.load(fullpath)
        if configpath is None:
            self.config = defaults.NNConfig()
        else:
            self.config = defaults.NNConfig(configpath)
        self.execute_info = {}

    @property
    def text(self):
        return str(lines_to_text(*self.lines))

    @text.setter
    def text(self, text):
        self.raw_lines = list(text_to_lines(text))
        # self.find_product()
        self.validate()
        self.load_variables()
        self.load_content()

    @property
    def raw_text(self):
        return str(lines_to_text(*self.raw_lines))

    @property
    def lines(self):
        new_lines = list(self.raw_lines)
        for ivar in self.variables.values():
            text = ivar.text
            idx = ivar.metadata['line_idx']
            new_lines[idx] = text
        return new_lines

    @property
    def default_command_args(self):
        return self.config.config[self.product]

    @property
    def configpath(self):
        return self.config.fullpath

    @property
    def filename_only(self):
        return get_filename(self.fullpath, ext=False)

    @filename_only.setter
    def filename_only(self, name):
        ext = get_file_extension(self.fullpath)
        self.filename = f'{name}{ext}'

    @property
    def filename(self):
        return get_filename(self.fullpath, ext=True)

    @filename.setter
    def filename(self, name):
        self.fullpath = os.path.join(self.folder_input, name)

    @property
    def folder_input(self):
        return get_folder(self.fullpath)

    @folder_input.setter
    def folder_input(self, folder):
        self.fullpath = os.path.join(folder, self.filename)

    @property
    def folder_output(self):
        key = 'outputdirectory'
        if key in self.execute_info.keys():
            return self.execute_info[key]
        else:
            raise KeyError('Input file has not been executed yet')

    def preview(self, nums=True):
        """
        Print the text of the file.

        Parameters
        ----------
        nums : bool, optional
            If it is True, it will show the number of each line. (default is True)
        """
        for i, line in enumerate(self.lines):
            if nums:
                print(f'{i} {line}')
            else:
                print(f'{line}')

    @load_message
    def load(self, fullpath):
        """
        The steps are the following:
            1. clear the current information
            2. load the raw text (update .fullpath and .raw_lines)
            3. find the nextnano product (update .product)
            4. validate the input file
            5. load the input variables (update .variables)
            6. load content (when applicable)

        Parameters
        ----------
        fullpath : str
            path to the file to be loaded
        """
        self.clear()
        self.fullpath = fullpath
        self.load_raw()
        self.find_product()
        self.validate()
        self.load_variables()
        self.load_content()

    def find_product(self):
        self.product = defaults.input_text_type(self.raw_text)
        return self.product

    def validate(self):
        if self.product not in defaults.products:
            self.product = 'Not valid'
            #raise ValueError(f'Not valid input file')

    @save_message
    def save(self, fullpath=None, overwrite=False, automkdir=True):
        """
        Save the current information into a file.

        Parameters
        ----------
        fullpath : str, optional
            path including the file name where it will be saved (default is None)
            If None, it will use the current .fullpath
        overwrite : bool, optional
            If it is False, it will find an unused name by adding an extra index like _1
            (default is False)
        automkdir : bool, optional
            If it is True, it will create the folder if it does not exist.
            (default is False)
        """
        if fullpath is None:
            if self.fullpath is None:
                raise ValueError('Please, specify a fullpath')
            fullpath = self.fullpath
        self.fullpath = savetxt(fullpath=fullpath, text=self.text, overwrite=overwrite, automkdir=automkdir)
        return self.fullpath

    @execute_message
    def execute(self, show_log = True, convergenceCheck = False, convergence_check_mode = 'pause',**kwargs):
        """
        Execute the input file located at .fullpath
        Individual kwargs can be passed like 'license' or 'database'
        If no kwargs is specified, it will use the default values in .config

        Parameters
        ----------
        show_log : bool, optional
            if False, suppress the simulation log
            (default is True)
        convergenceCheck : bool, optional
            if True, check convergence of the simulation
            (default is False)
        convergence_check_mode: string
            works only for convergenceCheck = True
            options:
                'pause': asks user how to proceed if simulation did not converge (default)
                'terminate': terminate the script if the simulation did not converge
                'continue': notify a user but continues execution of script
        kwargs may contain:
            exe : str, optional
                path to executable
            license : str, optional
                path to license file
            database : str, optional
                path to database file
            outputdirectory : str, optional
                path where to save the simulated data
        and other parameters depending on the nextnano product.
        For example, 'threads' is accepted by nextnano++.
        See the documentation of the command line arguments of each nextnano product
        on the online Manual (https://www.nextnano.de/manual/).
        """

        cmd_kwargs = dict(self.default_command_args)
        cmd_kwargs.update(kwargs)
        cmd_kwargs['inputfile'] = self.fullpath
        info = cmd_execute(show_log=show_log, parallel = self.__parallel__, **cmd_kwargs)
        self.execute_info = info
        if convergenceCheck and not self.__parallel__: #- possible solution for later (delete comment if not)
            self.check_convergence(mode= convergence_check_mode)
        return info

    def check_convergence(self, mode = 'pause'):
        if self.product == 'nextnano.MSB':
            raise NotImplementedError('Convergence check has not yet implemented for nextnano.MSB!')
            
        log = self.execute_info['logfile']
        try:
            if self.product == 'nextnano3':
                with open(log, 'r') as file:
                    for line in file:
                        if 'Exiting iteration and terminating simulation' in line or 'Program was terminated using a soft kill' in line or 'Terminating immediately' in line:
                            raise RuntimeError(f'\nSimulation got terminated! Check the log:\n{log}')
                        elif 'Maximum number of iterations exceeded' in line:
                            raise RuntimeError(f'\nSimulation did not converge! Check the log:\n{log}')
            elif self.product == 'nextnano++':
                with open(log, 'r') as file:
                    for line in file:
                        if 'Terminating program' in line:
                            raise RuntimeError(f'\nSimulation got terminated! Check the log:\n{log}')
                        elif 'Maximum number of iterations exceeded' in line:
                            raise RuntimeError(f'\nSimulation did not converge! Check the log:\n{log}')
                        elif 'Outdated numerics library (f95library) used' in line:
                            raise RuntimeError(f'\nOutdated numerics library (f95library) used.')
            elif self.product == 'nextnano.NEGF' or self.product == 'nextnano.NEGF++':   # NEGF reports convergence at every voltage and temperature sweep.
                with open(log, 'r') as file:
                    for line in file:
                        if 'Simulation has NOT CONVERGED' in line:
                            raise RuntimeError(f'\nSimulation has diverged! Check the log:\n{log}')
                        elif 'Simulation has partially converged' in line:
                            print(f'\nWARNING: check_convergence(): Simulation did not fully converge.')
                        elif 'Terminating program!' in line:
                            raise RuntimeError(f'\nSimulation got terminated! Check the log:\n{log}')
        except FileNotFoundError:
            print(f'Log file {log} not found!')
            if mode != 'continue':
                raise
            else:
                pass
        except RuntimeError as e:
            print(e)
            if mode == 'pause':
                pause = True
                while pause:
                    answer = input('Do you nevertheless want to continue? [y/n]: ')
                    if answer == 'y' or answer == 'yes':
                        pause = False
                        return
                    elif answer == 'n' or answer == 'no':
                        pause = False
                        raise RuntimeError('Nextnanopy terminated.')
                    else:
                        print('Invalid input.')
                        continue
            elif mode == 'terminate':
                raise RuntimeError('Nextnanopy terminated.')
            elif mode == 'continue':
                return
            else:
                raise ValueError(f'Mode "{mode}" is not valid')
        else:
            return

    def clear(self):
        self.raw_lines = []
        self.variables = DictList()
        self.fullpath = None

    def remove(self):
        os.remove(self.fullpath)

    def load_raw(self):
        with open(self.fullpath, 'r') as f:
            text = f.read()
        self.raw_lines = list(text_to_lines(text))
        return self.raw_lines

    def load_variables(self):
        pass

    def load_content(self):
        pass

    def get_variable(self, name):
        """
        Equivalent to self.variables[name]

        Parameters
        ----------
        name : str
            key for self.variables

        Raises
        ------
        KeyError
            If name is not a key of self.variables
        """
        if name not in self.variables.keys():
            raise KeyError(f'{name} is not a valid variable.')
        return self.variables[name]

    def set_variable(self, name, value=None, comment=None, unit = None):
        """
        Change the value and/or the comment of self.variable[name]

        Parameters
        ----------
        name : str
            key for self.variables
        value : not defined, optional
            Equivalent to self.variables[name].value = value (default is None)
            If it is None, it won't change it
        comment : not defined, optional
            Equivalent to self.variables[name].comment = comment (default is None)
            If it is None, it won't change it
        unit : not defined, optional
            Equivalent to self.variables[name].unit = unit (default is None)
            If it is None, it won't change it
        """

        var = self.get_variable(name)
        if value is not None:
            var.value = value
        if comment is not None:
            var.comment = comment
        if unit is not None:
            var.unit  = unit
        return var

    def __getitem__(self, item):
        return self.variables[item]

    def __setitem__(self, item, value):
        self.variables[item] = value

    def __delitem__(self, item):
        del self.variables[item]

    def __repr__(self):
        out = []
        out.append(f'{self.__class__.__name__}')
        out.append(f'fullpath: {self.fullpath}')
        out.append(f'Input variables: {len(self.variables)} elements')
        for key, var in self.variables.items():
            out.append(f'\t{str(var)}')
        out = '\n'.join(out)
        return out

    def __iter__(self):
        self._iter_index = 0
        return self

    def __next__(self):
        try:
            result = self.variables.__getitem__(self._iter_index)
        except (IndexError, KeyError):
            raise StopIteration
        self._iter_index += 1
        return result


# def decor(func):
#     def inner(*args,**kwargs):
#
#         args = list(args)
#         initial_object = args[0]
#         if hasattr(initial_object, '_helper_input_file'):
#             helper =  initial_object._helper_input_file
#         else:
#             func(*args,**kwargs)
#             return
#
#         helper_func = getattr(helper, func.__name__)
#
#
#         args = args[1:]
#
#
#         # print(kwargs)
#         #print('Start of decor message')
#         helper_func(*args, **kwargs)
#
#         for key in helper.__dict__:
#             setattr(initial_object, key, helper.__dict__[key])
#         #print('End of decor message')
#     return inner
#
# def class_decorator(cls):
#     for attr_name in dir(cls):
#         attr_value = getattr(cls, attr_name)
#         if hasattr(attr_value, '__call__') and not attr_name.startswith('__') and attr_name !='load': # check if attr is a function
#             setattr(cls, attr_name, decor(attr_value))
#     return cls



#@class_decorator
class InputFile(InputFileTemplate):
    # def __init__(self,fullpath = None, configpath = None):
    #
    #     self.raw_lines = []
    #     self.variables = DictList()
    #     self.content = None
    #     self.fullpath = fullpath
    #     self.product = 'not valid'
    #     self.__parallel__ = False
    #     if fullpath is not None:
    #         self.clear()
    #         self.fullpath = fullpath
    #         self.load_raw()
    #         self.find_product()
    #         self.validate()
    #     if configpath is None:
    #         self.config = defaults.NNConfig()
    #     else:
    #         self.config = defaults.NNConfig(configpath)
    #     self.execute_info = {}
    #
    #
    #     _helper_input_file_type = defaults.get_InputFile(self.product)
    #     self._helper_input_file = _helper_input_file_type(fullpath = fullpath, configpath = configpath)
    #
    #     for key in self._helper_input_file.__dict__:
    #         setattr(self, key, self._helper_input_file.__dict__[key])

    def __new__(cls, fullpath=None, configpath=None, *args, **kwargs):
        preInputFile = InputFileTemplate(fullpath,configpath)
        _InputFileType = defaults.get_InputFile(preInputFile.product)
        return _InputFileType(fullpath, configpath)

    # def load_variables(self):
    #     """
    #     Convenient method to find the best loading method for each nextnano product.
    #     self.variables will be updated
    #     """
    #     _InputFile = defaults.get_InputFile(self.product)
    #     file = _InputFile()
    #     file.text = self.text
    #     self.variables = file.variables
    #
    # def load_content(self):
    #     _InputFile = defaults.get_InputFile(self.product)
    #     file = _InputFile()
    #     file.text = self.text
    #     file.load_content()
    #     self.content = file.content
    #
    # @property
    # def lines(self):
    #     """
    #     update method to support NEGF input
    #     """
    #     try:
    #         _InputFile = defaults.get_InputFile(self.product)
    #     except ValueError:
    #         _InputFile = InputFileTemplate
    #
    #
    #     file = _InputFile()
    #     #file.product = self.product
    #     #file.text = self.raw_text
    #     #file.variables = self.variables
    #     #print(file.product)
    #     #print(type(file))
    #     file.variables = self.variables
    #     file.raw_lines = self.raw_lines
    #     return file.lines

class ExecutionQueue(threading.Thread):
    """
        This class take InputFiles and add them in the execution queue.
        Depending on limit_parallel, InputFiles are executed in parallel or sequentially.


        Parameters
        ----------
        limit_parallel: int
            number of InputFiles to be executed in parallel (default: 1)
        terminate_empty : bool
            If True, terminates once all added files are executed and logged.
            If you want to add more input files even after execution of all added in the beginning, use termanate_empy = False
            Then the ExecutionQueue has to be stopped manually later (ExecutionQueue.stop())
        convergenceCheck: bool
            see convergenceCheck in InputFile

        **execution_kwargs: parameters to be taken by InputFile.execute()




        Attributes
        ----------
        waiting_queue: queue.Queue
            queue of InputFile objects to be executed
        started: list
            list of (simulation_info:dict, InputFile) currently executing

        finished: list
            list of simulation_infos for finished simulations

        stop_when_empty: bool
            see terminate_empty parameter
        daemon: bool
            see threading.Thread.daemon

        Methods
        ----------- for user
        add(*input_files)
            adds InputFiles to queue

        start()
            start the thread (i.e. execution)
            see threading.Thread.start()

        stop()
            stop the thread (once all added files are executed)
            only necessary if termanate_empty = True


        -------internal (or for advanced users)
        all_done()
            return True if all execution and logging are finished

        add_execution()
            pop an InputFile from self.waiting_queue, execute and add to self.started

        log_finished()
            finish logging for finished execution in self.started

        run()
            commands to be run upon start():
                pop simulation from queue and execute
                log the simulation if some are finished from self.started
            see threading.Thread.run()

    """
    def __init__(self, limit_parallel : int = 1 , maxsize : int = 0, terminate_empty : bool = True, convergenceCheck = False, **execution_kwargs):
        super(ExecutionQueue, self).__init__()
        self.waiting_queue = queue.Queue()#should be queue of InputFile objects
        self.started = []#should be list of execution_infos
        self.finished = []#should be list of execution_infos
        self.limit_parallel = limit_parallel
        self.execution_kwargs = execution_kwargs
        self.convergenceCheck = convergenceCheck
        self.stop_when_empty = terminate_empty
        self.daemon = False
    def all_done(self):
        return not bool(not self.waiting_queue.empty() or self.started)

    def add(self,*input_files: InputFileTemplate):
        for input_file in input_files:
            self.waiting_queue.put(input_file)

    def stop(self):
        self.stop_when_empty = True


    def add_execution(self):
        if (len(self.started) < self.limit_parallel) and not self.waiting_queue.empty():
            input_f = self.waiting_queue.get()
            if self.limit_parallel>1:
                input_f.__parallel__ = True
            if 'show_log' in self.execution_kwargs and not self.execution_kwargs['show_log']:
                print(f"\nRemaining simulations in the queue: ", self.waiting_queue.qsize())
            info = input_f.execute(**self.execution_kwargs)
            self.started.append((info,input_f))

    def log_finished(self):
        if self.limit_parallel>1:
            i = 0
            while i < len(self.started):
                poll = self.started[i][0]['process'].poll()
                if poll is None:
                    i+=1
                else:
                    #TODO check once again del
                    self.started[i][0]['process'].wait()
                    tout = self.started[i][0]['tout']
                    terr = self.started[i][0]['terr']
                    for t in (tout, terr):
                        t.join()
                    self.started[i][0]['queue'].put(None)
                    if self.convergenceCheck:
                        if 'convergence_check_mode' in self.execution_kwargs:
                            convergence_check_mode = self.execution_kwargs['convergence_check_mode']
                        else:
                            convergence_check_mode = 'pause'
                        self.started[i][1].check_convergence(mode = convergence_check_mode)
                    self.finished.append(self.started[i][0])
                    del self.started[i]


        else:
            i = 0
            while i < len(self.started):
                poll = self.started[i][0]['process'].poll()
                if poll is None:
                    i+=1
                else:
                    self.finished.append(self.started[i][0])
                    del self.started[i]

    def run(self):
        while True:
            self.add_execution()
            self.log_finished()
            if self.all_done():
                print('\nWaiting queue is empty, all execution and logging are finished')

                if self.stop_when_empty:
                    break
                while self.all_done() and not self.stop_when_empty:
                    time.sleep(0.1) #to ensure switch to main thread
                    pass



class Sweep(InputFileTemplate):
    """
        This class give a user possibility to run multiple simulations (sweep) over defined variables in the input file.


        Parameters:
        -------------------
        variables_to_sweep: dict
            Dict of variables to sweep in the form of {name1:values1,name2:values2...}
            values should be an iterable object (ideally list)
        fullpath: str
            defined as for InputFile
        configpath: str
            defined as for input files


        Methods:
        --------
        save_sweep()
            creates an output folder
            creates input files for all combinations of sweep variables
        execute_sweep():
            execute created input files and saves information to output folder

    """
    def __init__(self,variables_to_sweep,fullpath=None, configpath=None):
        testInputFile = InputFile(fullpath = fullpath, configpath = configpath)
        super().__init__(fullpath, configpath)
        if set(variables_to_sweep.keys()).issubset(testInputFile.variables.keys()):
            self.var_sweep = variables_to_sweep
        else:
            raise ValueError('Defined variables are not variables of input file')
        for value in self.var_sweep.values():
            if not isinstance(value, Iterable):
                raise TypeError('Values of variables_to_sweep should be iterable objects')
        self.sweep_output_directory = None
        self.input_files = []
        self.sweep_infodict = DictList()
        self.sweep_output_infodict = DictList()

    def save_sweep(self, delete_old_files = True, round_decimal = 8, integer_only_in_name = False):
        """

        Parameters
        ----------
        delete_old_files: if True, deletes files created in previous sweeps
        round: number of digits to round in the output folder names

        Returns
        -------
        None
        """
        if delete_old_files == True:
            for inputfile in self.input_files:
                inputfile.remove()
        self.input_files = []
        self.create_input_files(round_decimal, integer_only_in_name = integer_only_in_name)

    def prepare_output(self, overwrite = False, output_directory = None):
        self.sweep_output_directory = self.mk_dir(overwrite=overwrite, output_directory = output_directory)
        self.create_info()

    def create_input_files(self, round_decimal, integer_only_in_name = False):
        iteration_combinations = list(itertools.product(*self.var_sweep.values()))
        filename_path, filename_extension = os.path.splitext(self.fullpath)
        for combination in iteration_combinations:
            filename_end = '__'
            inputfile = InputFile(fullpath = self.fullpath, configpath = self.configpath)
            for var_name, var_value in zip(self.var_sweep.keys(), combination):
                inputfile.set_variable(var_name, var_value, comment='THIS VARIABLE IS UNDER SWEEP')
                if isinstance(var_value,str):
                    var_value_string = var_value
                else:
                    var_value_string = round(var_value, round_decimal)
                filename_end += '{}_{}_'.format(var_name, var_value_string)
            if integer_only_in_name:
                inputfile.save(overwrite = False)
            else:
                inputfile.save(filename_path + filename_end + filename_extension, overwrite = True)
            variable_combination =  dict(zip(self.var_sweep.keys(), combination))
            self.input_files.append(inputfile)
            self.sweep_infodict[inputfile.fullpath] = variable_combination


    def execute_sweep(self, delete_input_files = False, overwrite = False, show_log = True, convergenceCheck = False, convergence_check_mode = 'pause', parallel_limit = 1, separate_sweep_dir = True, **kwargs):
        """
        Execute created input files and saves information to output folder.

        Parameters
        ----------
        delete_input_files : bool, optional
            if True, input_files are deleted after execution. Default is False.
        overwrite : bool, optional
            if True, the output overwrites the old output data. If False, execution will create a new output folder
            (with the unique name, created by adding an integer to the foldername). Default is False.
        show_log : bool, optional
            if True, the simulation log is displayed in the console. If False, the count of current simulation is displayed without log. 
            Default is True.
            Note that the log file is always saved in the output folders regardless of this option.
        convergenceCheck : bool, optional
            if True, nextnanopy scans the log file of the simulation performed and check whether the solution has converged. 
            If it did not converge, nextnanopy warns you and ask if you want to proceed with postprocessing. 
            Note that non-converged solutions are not reliable and further calculation and/or visualization from them do not make much sense. 
            Default is False.
        parallel_limit: int, optional
            number of simulation to run simultaniously. Espicially usefull for simple simulations which migh be more efficiently rn in parallel. Be aware that
            some nextnano solvers parallelize computations internally in threads (controlled by --threads in nextnanopy config). To avoid unexpected behaviour and
            not desirable decrease of simulation speed use the rule: parallel_limit*threads<= number of physical cores of the mahcine
            default 1
        separate_sweep_dir: bool, optional
            if True, creates separate directory to store subdirectories of the sweep simulation. If False, stores all directories without separate directory.
            default True
        **kwargs:
            see **kwargs of InputFile.execute()
        """
        try:
            output_directory = kwargs['outputdirectory']
            del kwargs['outputdirectory']
        except KeyError:
            output_directory = self.config.get(section = self.product,option = 'outputdirectory')
        if separate_sweep_dir:
            self.prepare_output(overwrite, output_directory)
            output_directory = self.sweep_output_directory
        else:
            self.sweep_output_directory = output_directory

        if not self.input_files:
            warnings.warn('Nothing was executed in sweep! Input files to execute were not created.')
            return

        #TODO: delete if statement (use execution_queue for both cases)
        if parallel_limit>1:
            execution_queue = ExecutionQueue(limit_parallel=parallel_limit, terminate_empty=True, outputdirectory = output_directory, show_log = show_log, convergenceCheck = convergenceCheck, convergence_check_mode = convergence_check_mode, **kwargs)
            execution_queue.add(*self.input_files)
            execution_queue.start()

            execution_queue.join()
            if delete_input_files:
                for inputfile in self.input_files:
                    inputfile.remove()
        else:
            for i, inputfile in enumerate(self.input_files):
                if not show_log:
                    print(f"\nExecuting simulations [{i+1}/{len(self.input_files)}]...")
                info = inputfile.execute(outputdirectory = output_directory, show_log = show_log, convergenceCheck = convergenceCheck, convergence_check_mode = convergence_check_mode,**kwargs)
                if delete_input_files:
                    inputfile.remove()

        #part where the info is stored
        for inputfile, variable_combination in zip(self.input_files, self.sweep_infodict.values()):
            self.sweep_output_infodict[inputfile.folder_output] = variable_combination
        # TODO create files with info in output_directories
        if True:# TODO
            # self.create_infodict_files()
            self.create_infodict_json()

    def create_infodict_files(self):
        """
        Creates files with variables under sweep in output directories

        """
        raise NotImplementedError

    def create_infodict_json(self):
        """
        Creates json file to store infodict
        """
        import json
        filepath = os.path.join(self.sweep_output_directory, 'sweep_infodict.json')
        with open(filepath, "w") as file:
            json.dump(self.sweep_output_infodict, file, indent=4)

    def mk_dir(self,overwrite = False, output_directory = None):
        vars = ''
        for i in self.var_sweep.keys():
            vars+=('__'+i)
        name_of_file = self.filename_only
        if not output_directory:
            output_directory = self.config.get(section = self.product,option = 'outputdirectory')
        name = (name_of_file+'_sweep'+vars)
        if overwrite:
            directory = mkdir_if_not_exist(os.path.join(output_directory,name))
        else:
            directory = mkdir_even_if_exists(output_directory,name)
        return directory

    def create_info(self):
        file_location = os.path.join(self.sweep_output_directory,'sweep_info.txt')
        with open(file_location,'w') as file:
            file.write("Input file: '{}' \n".format(self.fullpath))
            file.write("Sweep variables: \n")
            for i in self.var_sweep:
                file.write("{} = {} \n".format(i,self.var_sweep[i]))







import unittest
from nextnanopy.utils.datasets import Data, Variable, Coord
from nextnanopy.utils.datasets import default_unit
import numpy as np


class TestDatasets(unittest.TestCase):
    def test_Data(self):
        ds = Data(name='test', value=2)
        self.assertEqual(ds.name, 'test')
        self.assertEqual(ds.value, np.array(2))
        self.assertEqual(ds.unit, default_unit)
        self.assertEqual(ds.metadata, {})

    def test_Variable(self):
        ds = Variable(name='test', value=2)
        self.assertEqual(ds.name, 'test')
        self.assertEqual(ds.value, np.array(2))
        self.assertEqual(ds.unit, default_unit)
        self.assertEqual(ds.metadata, {})

    def test_Coord(self):
        ds = Coord(name='test', value=2, dim=3, offset=10)
        self.assertEqual(ds.name, 'test')
        self.assertEqual(ds.value, np.array(2))
        self.assertEqual(ds.unit, default_unit)
        self.assertEqual(ds.metadata, {})
        self.assertEqual(ds.dim, 3)
        self.assertEqual(ds.offset, np.array(10))
        self.assertEqual(ds.get_value(use_offset=True), ds.value + ds.offset)
        self.assertEqual(ds.get_value(use_offset=False), ds.value)


if __name__ == '__main__':
    unittest.main()



from nextnanopy.utils.formatting import text, lines_to_text, modify_lines_in_text, paragraph, text_to_lines, preview

cbkl = "{"
cbkr = "}"
sbkl = "["
sbkr = "]"
quo = '"'
input_extension = ".in"


def sbk(init='', mid='', end=''):
    return sbkl + text(init, mid, end) + sbkr


def cbk(init='', mid='', end=''):
    return cbkl + text(init, mid, end) + cbkr


def quote(init='', mid='', end=''):
    return quo + text(init, mid, end) + quo


class InputAssistant(object):
    rc_default = {
        'indent': ' ' * 3,
        'new_line': '\n',
        'equal': ' = ',
        'block_text_init': ' ',
        'block_text_end': ' ',
        'block_init': '',
        'block_end': '',
        'comment': '# ',
        'oif': '#if ',
        'nif': '!IF',
        'when': '!WHEN ',
        'elif': '!ELIF',
        'else': '!ELSE',
        'endif': '!ENDIF',
    }

    def __init__(self):
        self.text = text
        self.sbk = sbk
        self.cbk = cbk
        self.lines = lines_to_text
        self.modify_lines = modify_lines_in_text
        self.paragraph = paragraph
        self.split_lines = text_to_lines
        self.quote = quote
        self.nline = '\n'
        self.preview = preview

    @property
    def rc(self):
        if hasattr(self, '_rc'):
            return self._rc
        else:
            return self.rc_default

    @rc.setter
    def rc(self, params):
        self._rc = rc

    # -- Basic methods
    def _indent(self, level):
        return self.get_indent_level(level)

    def get_indent_level(self, level):
        return self.rc['indent'] * level

    def equal(self, name, value):
        eq = self.rc['equal']
        return f'{name}{eq}{value}'

    def point(self, name, *values):
        vector = ', '.join([str(vi) for vi in values])
        vector = self.sbk(vector)
        return self.equal(name, vector)

    def block(self, name, content):
        bi, be = self.rc['block_init'], self.rc['block_end']
        ti, te = self.rc['block_text_init'], self.rc['block_text_end']
        return f'{name}{bi}{self.cbk(ti, content, te)}{be}'

    def add_indent(self, text, amount=1):
        return self.modify_lines(self.rc['indent'] * amount, text)

    # -- Specific methods
    def equal_lines(self, **kwargs):
        content = [self.equal(key, value) for key, value in kwargs.items()]
        return self.lines(*content)

    def equal_block(self, name, equal_kw):
        content = self.equal_lines(**equal_kw)
        content = self.add_indent(content)
        content = self.paragraph(content)
        return self.block(name, content)

    def merge_blocks(self, name, *blocks):
        content = self.lines(*blocks)
        content = self.add_indent(content)
        content = self.paragraph(content)
        return self.block(name, content)

    def comment_lines(self, text):
        return self.modify_lines(self.rc['comment'], text)

    def comment(self, text, end='\n'):
        return self.text(self.rc['comment'], text, end)

    def when_line(self, text, variable):
        init = self.rc['when'] + f"${variable} "
        return self.modify_lines(init, text)

    def if_lines(self, text, variable):
        init = self.rc['nif'] + f'${variable} '
        return self.modify_lines(init, text)

    def if_block(self, if_variable, if_content, elif_variable=None,
                 elif_content=None, else_content=None):
        content = (self.rc['nif'] + f'(${if_variable})' + self.rc['new_line']
                   + self.modify_lines(self.rc['indent'], if_content)
                   + self.rc['new_line'])
        if elif_variable is not None:
            content += (self.rc['elif'] + f'(${elif_variable})'
                        + self.rc['new_line']
                        + self.modify_lines(self.rc['indent'], elif_content)
                        + self.rc['new_line'])
        if else_content is not None:
            content += (self.rc['else'] + f'(${else_variable})'
                        + self.rc['new_line']
                        + self.modify_lines(self.rc['indent'], else_content)
                        + self.rc['new_line'])
        content += self.rc['endif']
        return content

    # - Variables
    def variables(self, **kwargs):
        content = self.equal_lines(**kwargs)
        return self.modify_lines('$', content)

    # - Global
    def global_simulate(self, dim=1):
        return self.block(f'simulate{int(dim)}D', '')

    def global_crystal(self, kind='zb', x_hkl=[1, 0, 0], y_hkl=[0, 1, 0]):
        points = [self.point(key, *value) for key, value in dict(x_hkl=x_hkl, y_hkl=y_hkl).items()]
        content = self.lines(*points)
        content = self.add_indent(content)
        content = self.paragraph(content)
        return self.block(f'crystal_{kind}', content)

    def global_substrate(self, name, alloy_x=None, alloy_y=None):
        kwargs = {'name': name}
        if alloy_x:
            kwargs['alloy_x'] = alloy_x
        if alloy_y:
            kwargs['alloy_y'] = alloy_y
        return self.equal_block('substrate', kwargs)

    def global_periodic(self, x=False, y=None, z=None):
        inputs = dict(x=x, y=y, z=z)
        kwargs = {}
        for key, value in inputs.items():
            value = str(value).lower()
            if value == 'none':
                continue
            elif value in ['true', 'yes', 'y']:
                value = 'yes'
            else:
                value = 'no'
            kwargs[key] = value
        return self.equal_block('periodic', kwargs)

    def global_temperature(self, value):
        return self.equal('temperature', value)

    def global_block(self, *blocks):
        return self.merge_blocks('global', *blocks)

    # - Grid
    def grid_line(self, pos, spacing):
        line = self.equal('pos', pos) + '\t' + self.equal('spacing', spacing)
        return self.block('line', line)

    def grid_axis(self, *pos_spacing, axis='x', min_pos=None, max_pos=None):
        content = [self.grid_line(pi, spi) for pi, spi in pos_spacing]
        for key, value in zip(['min_pos', 'max_pos'], [min_pos, max_pos]):
            if value:
                content.append(self.equal(key, value))
        content = self.lines(*content)
        return self.merge_blocks(f'{axis}grid', content)

    def grid_x(self, *pos_spacing, min_pos=None, max_pos=None):
        return self.grid_axis(*pos_spacing, min_pos=min_pos, max_pos=max_pos, axis='x')

    def grid_y(self, *pos_spacing, min_pos=None, max_pos=None):
        return self.grid_axis(*pos_spacing, min_pos=min_pos, max_pos=max_pos, axis='y')

    def grid_z(self, *pos_spacing, min_pos=None, max_pos=None):
        return self.grid_axis(*pos_spacing, min_pos=min_pos, max_pos=max_pos, axis='z')

    def grid_block(self, *blocks):
        return self.merge_blocks('grid', *blocks)

    # - Structure
    def structure_output(self, name, boxes=None):
        boxes = str(boxes).lower()
        if boxes == 'none':
            return self.block(name, '')
        if boxes in ['true', 'yes', 'y']:
            value = 'yes'
        else:
            value = 'no'
        return self.block(name, self.equal('boxes', value))

    def structure_output_region_index(self, boxes=None):
        return self.structure_output('output_region_index', boxes)

    def structure_output_material_index(self, boxes=None):
        return self.structure_output('output_material_index', boxes)

    def structure_output_contact_index(self, boxes=None):
        return self.structure_output('output_contact_index', boxes)

    def structure_output_allow_composition(self, boxes=None):
        return self.structure_output('output_allow_composition', boxes)

    def structure_output_impurities(self, boxes=None):
        return self.structure_output('output_impurities', boxes)

    def structure_output_generation(self, boxes=None):
        return self.structure_output('output_generation', boxes)

    def structure_block(self, *blocks):
        return self.merge_blocks('structure', *blocks)

    # - Region
    def region_everywhere(self):
        return self.block('everywhere', '')

    def region_contact(self, name):
        content = self.equal('name', name)
        return self.block('contact', content)

    def region_binary(self, name):
        content = self.equal('name', name)
        return self.block('binary', content)

    def region_ternary_constant(self, name, alloy_x):
        name = self.quote(name)
        return self.equal_block('ternary_constant', dict(name=name, alloy_x=alloy_x))

    def region_line(self, x):
        content = self.point('x', *x)
        return self.block('line', content)

    def region_rectangle(self, x, y):
        points = [self.point(key, *value) for key, value in dict(x=x, y=y).items()]
        content = self.lines(*points)
        content = self.add_indent(content)
        content = self.paragraph(content)
        return self.block('rectangle', content)

    def region_obelisk(self, base_x, base_y, base_z, top_x, top_y, top_z):
        kwargs = dict(
            base_x=base_x, base_y=base_y, base_z=base_z,
            top_x=top_x, top_y=top_y, top_z=top_z)
        points = [self.point(key, *value) for key, value in kwargs.items()]
        content = self.lines(*points)
        content = self.add_indent(content)
        content = self.paragraph(content)
        return self.block('obelisk', content)

    def region_cuboid(self, x, y, z):
        kwargs = dict(x=x, y=y, z=z)
        points = [self.point(key, *value) for key, value in kwargs.items()]
        content = self.lines(*points)
        content = self.add_indent(content)
        content = self.paragraph(content)
        return self.block('cuboid', content)

    def region_polygonal_prism(self, axes=['x', 'y', 'z'], vertexes=[[10.5, 14.0]], height=[0, 10]):
        point = self.point(axes[2], *height)
        blocks = [self.region_vertex(ax1=axes[0], ax2=axes[1], value1=v[0], value2=v[1]) for v in vertexes]
        return self.merge_blocks('polygonal_prism', point, *blocks)

    def region_vertex(self, ax1='x', ax2='y', value1=1.0, value2=2.0):
        kwargs = {f'{ax1}': value1, f'{ax2}': value2}
        points = [self.point(key, value) for key, value in kwargs.items()]
        content = self.lines(*points)
        content = self.add_indent(content)
        content = self.paragraph(content)
        return self.block(f'vertex', content)

    def region_doping_constant(self, name, conc):
        cblock = self.equal_block('constant', dict(name=name, conc=conc))
        return self.merge_blocks('doping', cblock)

    def region_doping_remove(self):
        cblock = self.block('remove','')
        return self.merge_blocks('doping', cblock)

    def region_integrate(self, label='', electron=False, hole=False, piezo=False, pyro=False, polarization=False):
        kwargs = {
            'electron_density': electron,
            'hole_density': hole,
            'piezo_density': piezo,
            'pyro_density': pyro,
            'polarization': polarization,
        }
        content = []
        for key, value in kwargs.items():
            if value:
                content.append(self.block(key, ''))
        if label:
            label = self.quote(label)
            content.append(self.equal('label', label))
        content = self.lines(*content)
        content = self.add_indent(content)
        content = self.paragraph(content)
        return self.block('integrate', content)

    def region_block(self, *blocks):
        return self.merge_blocks('region', *blocks)

    # - Impurities
    def impurities_donor(self, name, energy, degeneracy):
        return self.equal_block('donor', dict(name=name, energy=energy, degeneracy=degeneracy))

    def impurities_acceptor(self, name, energy, degeneracy):
        return self.equal_block('acceptor', dict(name=name, energy=energy, degeneracy=degeneracy))

    def impurities_charge(self, name, positive=True):
        kind = 'positive' if positive else 'negative'
        return self.equal_block('charge', dict(name=name, type=kind))

    def impurities_block(self, *blocks):
        return self.merge_blocks('impurities', *blocks)

    # - Contacts
    def contacts_schottky(self, name, bias, barrier,steps = 1):
        if bias == []:
            raise ValueError('bias could be a number or non-zero length list of numbers')
        if steps == 1:
            return self.equal_block('schottky', dict(name=name, bias=bias, barrier=barrier))
        else:
            return self.equal_block('schottky', dict(name=name, bias=bias, barrier=barrier,steps = steps))

    def contacts_fermi(self, name, bias,steps = 1):
        if bias == []:
            raise ValueError('bias could be a number or non-zero length list of numbers')
        return self.equal_block('fermi', dict(name=name, bias=bias,steps = steps))

    def contacts_ohmic(self, name, bias, shift=0,steps = 1):
        if bias == []:
            raise ValueError('bias could be a number or non-zero length list of numbers')
        return self.equal_block('ohmic', dict(name=name, bias=bias, shift = shift,steps = steps))

    def contacts_block(self, *blocks):
        return self.merge_blocks('contacts', *blocks)

    # - Classical
    def classical_gamma(self, output_bandedges=None):
        name = 'Gamma'
        if output_bandedges:
            return self.merge_blocks(name, output_bandedges)
        else:
            return self.block(name, '')

    def classical_hh(self, output_bandedges=None):
        name = 'HH'
        if output_bandedges:
            return self.merge_blocks(name, output_bandedges)
        else:
            return self.block(name, '')

    def classical_lh(self, output_bandedges=None):
        name = 'LH'
        if output_bandedges:
            return self.merge_blocks(name, output_bandedges)
        else:
            return self.block(name, '')

    def classical_so(self, output_bandedges=None):
        name = 'SO'
        if output_bandedges:
            return self.merge_blocks(name, output_bandedges)
        else:
            return self.block(name, '')

    def classical_bands(self):
        gamma = self.classical_gamma()
        hh = self.classical_hh()
        lh = self.classical_lh()
        so = self.classical_so()
        return self.lines(gamma, hh, lh, so)

    def classical_output_bandedges(self, averaged='yes', profiles=['Gamma', 'electron_fermi_level']):
        profiles = ' '.join(profiles)
        profiles = f'"{profiles}"'
        return self.equal_block('output_bandedges', dict(averaged=averaged, profiles=profiles))

    def classical_output_carrier_densities(self):
        return self.block('output_carrier_densities', '')

    def classical_output_intrinsic_density(self):
        return self.block('output_intrinsic_density', '')

    def classical_output_ionized_dopant_densities(self):
        return self.block('output_ionized_dopant_densities', '')

    def classical_block(self, *blocks):
        return self.merge_blocks('classical', *blocks)

    # - Poisson
    def poisson_output_potential(self):
        return self.block('output_potential', '')

    def poisson_output_electric_field(self):
        return self.block('output_electric_field', '')

    def poisson_newton_solver(self, iterations=30, search_steps=40, residual=1e-4):
        return self.equal_block('newton_solver',
                                dict(iterations=iterations, search_steps=search_steps, residual=residual))

    def poisson_debuglevel(self, value):
        return self.equal('debuglevel', value)

    def poisson_block(self, *blocks):
        return self.merge_blocks('poisson', *blocks)

    # - Currents
    def currents_output_mobilities(self):
        return self.block('output_mobilities', '')

    def currents_recombination_model(self, SRH='no', Auger='no', radiative='no', enable_generation='yes'):
        inputs = dict(SRH=SRH, Auger=Auger, radiative=radiative, enable_generation=enable_generation)
        kwargs = {}
        for key, value in inputs.items():
            if value:
                kwargs[key] = value
        return self.equal_block('recombination_model', kwargs)

    def currents_block(self, *blocks):
        return self.merge_blocks('currents', *blocks)

    # - Quantum
    def quantum_name(self, name):
        return self.equal('name', name)

    def quantum_coords(self, x, y=None, z=None):
        inputs = dict(x=x, y=y, z=z)
        kwargs = {}
        for key, value in inputs.items():
            if value:
                kwargs[key] = value
        points = [self.point(key, *value) for key, value in kwargs.items()]
        return self.lines(*points)

    def quantum_density(self, use=True):
        if use:
            value = 'no'
        else:
            value = 'yes'
        return self.equal('no_density', value)

    def quantum_boundary(self, x='neumann', y=None, z=None):
        inputs = dict(x=x, y=y, z=z)
        kwargs = {}
        for key, value in inputs.items():
            if value:
                kwargs[key] = value
        return self.equal_block('boundary', kwargs)

    def quantum_output_wavefunctions(self, max_num=5, amplitudes='no', probabilities='yes', all_k_points='yes'):
        return self.equal_block('output_wavefunctions',
                                dict(max_num=max_num, amplitudes=amplitudes, probabilities=probabilities,
                                     all_k_points=all_k_points))

    def quantum_gamma(self, num_ev=100):
        content = self.equal('num_ev', num_ev)
        return self.block('Gamma', content)

    def quantum_quantize_x(self):
        return self.block('quantize_x', '')

    def quantum_quantize_y(self):
        return self.block('quantize_y', '')

    def quantum_quantize_z(self):
        return self.block('quantize_z', '')

    def quantum_block(self, *blocks):
        region = self.merge_blocks('region', *blocks)
        return self.merge_blocks('quantum', region)

    # - Output
    def output_directory(self, path='output'):
        return self.equal('directory', path)

    def output_format(self, dim, value):
        return self.equal(f'format{dim}D', value)

    def output_section(self, name, dim=2, x=None, y=None, z=None, range_x=None, range_y=None, range_z=None):
        coords = dict(name=name, x=x, y=y, z=z)
        kwargs = {}
        for key, value in coords.items():
            if value is not None:
                kwargs[key] = value
        # dim = int(3 - (len(kwargs.keys()) - 1))

        coords = self.equal_lines(**kwargs)
        ranges = dict(range_x=range_x, range_y=range_y, range_z=range_z)
        kwargs = {}
        for key, value in kwargs.items():
            if value:
                kwargs[key] = value
        points = [self.point(key, value) for key, value in kwargs.items()]
        content = self.lines(coords, *points)
        content = self.add_indent(content)
        content = self.paragraph(content)
        return self.block(f'section{dim}D', content)

    def output_only_sections(self, value=False):
        value = str(value).lower()
        if value in ['true', 'yes', 'y']:
            value = 'yes'
        else:
            value = 'no'
        return self.equal('only_sections', value)

    def output_block(self, *blocks):
        return self.merge_blocks('output', *blocks)

    # - Run
    def run_solve_strain(self):
        return self.block('solve_strain', '')

    def run_solve_poisson(self):
        return self.block('solve_poisson', '')

    def run_solve_quantum(self):
        return self.block('solve_quantum', '')

    def run_outer_iteration(self, **kwargs):
        if not kwargs:
            return self.block('outer_iteration', '')
        else:
            return self.equal_block('outer_iteration', kwargs)

    def run_block(self, *blocks):
        return self.merge_blocks('run', *blocks)



from nextnanopy.utils.datasets import InputVariable
from nextnanopy.utils.formatting import str_to_path, _path, pattern_in_file, is_variable, parse_variable, \
    generate_command, pattern_in_text
from collections import OrderedDict


fmt = {
    'var_char': '$',
    'com_char': '<!--',
    'input_pattern': '<Simulation',
}

fmtpp = {
    'var_char': '$',
    'com_char': '#',
    'input_pattern': 'nextnano.NEGF{',
}

config_validator = {
    'exe': str_to_path,
    'license': str_to_path,
    'database': str_to_path,
    'outputdirectory': str_to_path,
    'threads': int,
}

config_default = {
    'exe': '',
    'license': '',
    'database': '',
    'outputdirectory': '',
    'threads': 0,
}


def command_negf(
        inputfile,
        exe,
        license,
        database,
        outputdirectory,
        threads=0,
        **kwargs,
):
    kwargs = OrderedDict(
        exe            = [_path(exe), ''],
        inputfile      = [_path(inputfile), ''],
        outputdirectory= [_path(outputdirectory), ''],
        database       = [_path(database), ''],
        license        = [_path(license), ''],
        threads        = ['-threads', threads],
    )
    return generate_command(kwargs.values())


def command_negfpp(
        inputfile,
        exe,
        license,
        database,
        outputdirectory,
        threads=0,
        **kwargs,
):
    kwargs = OrderedDict(
        exe            = [_path(exe), ''],
        inputfile      = ['--input-file', _path(inputfile)],
        outputdirectory= ['--outputdirectory', _path(outputdirectory)],
        database       = ['--database', _path(database)],
        license        = ['--license', _path(license)],
        threads        = ['--threads', threads],
    )
    return generate_command(kwargs.values())


def is_negf_variable(text):
    return is_variable(text, var_char=fmt['var_char'])


#def parse_negf_variable(text):
#    return parse_variable(text, var_char=fmt['var_char'], com_char=fmt['com_char'])

def parse_negf_variable_name(text):
    if text.startswith(fmt['var_char']):
        return text[1:]
    else:
        raise ValueError(f"variable name in NEGF should start with {fmt['var_char']}")

def is_negf_input_file(fullpath):
    return pattern_in_file(fullpath, fmt['input_pattern'])


def is_negfpp_input_file(fullpath):
    return pattern_in_file(fullpath, fmtpp['input_pattern'])


def is_negf_input_text(text):
    return pattern_in_text(text, fmt['input_pattern'])


def is_negfpp_input_text(text):
    return pattern_in_text(text, fmtpp['input_pattern'])


class InputVariable_NEGF(InputVariable):
    var_char = fmt['var_char']

    @property
    def text(self):
        t = f'{self.var_char}{self.name} = {self.value} {self.unit}'
        if self.comment:
            t = f'{t} Comment: {self.comment}'
        return t



import numpy as np
import matplotlib.pyplot as plt
from cycler import cycler
import gdspy
import warnings

units_factor = {
    'nm': 1e-9,
    'um': 1e-6,
    'mm': 1e-3,
    'm': 1,
    'si': 1,
}


def validate_unit(key):
    key = str(key).lower()
    if key not in units_factor.keys():
        raise KeyError(f'Invalid unit: {key}')
    return True


class GdsPolygonsRaw(object):
    def __init__(self, fullpath, unit='nm', **kwargs):
        self.fullpath = fullpath
        self._labels = []
        self._unit = 'm'
        self.load(**kwargs)
        self.unit = unit
        self.added_labels = []

    @property
    def xy(self):
        _xy = [np.array(pol_xy).transpose() for pol_xy in self.polygons_xy]
        return _xy

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, key):
        validate_unit(key)
        self.polygons_xy = [xy * units_factor[self._unit] / units_factor[key]
                            for xy in self.polygons_xy]
        self._unit = key

    @property
    def labels(self):
        return self._labels

    @labels.setter
    def labels(self, names):
        if len(names) == self.nb_polygons:
            self._labels = names
        else:
            print(f'Number of label ({names}) must be equal to number of polygons ({self.nb_polygons})')
            print(f'Using default labels: {self.labels}')

    @property
    def nb_polygons(self):
        return len(self.polygons_xy)

    def load(self, **kwargs):
        self.load_polygons_xy(**kwargs)
    @property
    def slices(self):
        warnings.warn("The GdsPolygonsRaw.slices is deprecated",DeprecationWarning)
        from nextnanopy.utils.shapes_deprecated import SlicedPolygon
        return [SlicedPolygon(pol_xy) for pol_xy in self.polygons_xy]

    def load(self):
        self.load_polygons_xy()
        self.set_default_labels()

    def load_polygons_xy(self, **kwargs):
        gds_lib = gdspy.GdsLibrary(infile=self.fullpath)
        xys = []
        if 'by_spec' not in kwargs:
            kwargs['by_spec'] = False
        for cell in gds_lib.top_level():
            pols = cell.get_polygons(kwargs['by_spec'])
            pols = [pi * gds_lib.unit for pi in pols]
            xys.extend(pols)
        self.polygons_xy = xys
        return xys

    def set_default_labels(self):
        if len(self.labels) == 0:
            self.labels = np.arange(self.nb_polygons)

    # -- Useful show methods
    def _prepare_ax(self, ax=None, cmap='nipy_spectral'):
        plt.ion()
        if not ax:
            fig, ax = plt.subplots(1)
        if cmap:
            colormap = plt.get_cmap(cmap)
            colors = [colormap(i) for i in np.linspace(0, 1, self.nb_polygons)]
            ax.set_prop_cycle(cycler('color', colors))
        ax.set_xlabel(f'x {self.unit}')
        ax.set_ylabel(f'y {self.unit}')
        return ax

    def show_all(self, ax=None, cmap='nipy_spectral', fill_kw={}):
        ax = self._prepare_ax(ax, cmap)
        for xy, label in zip(self.xy, self.labels):
            x, y = xy
            ax.fill(x, y, label=label, **fill_kw)
        ax.legend(loc='upper right')
        return ax

    def show_onebyone(self, ax=None, cmap='nipy_spectral', fill_kw={}):
        ax = self._prepare_ax(ax, cmap)
        for pol, label in zip(self.polygons, self.labels):
            x, y = pol.boundary.xy
            ax.fill(x, y, color='grey', alpha=0.5)
        for pol, label in zip(self.polygons, self.labels):
            x, y = pol.boundary.xy
            ax.fill(x, y, label=label, **fill_kw)
            ax.legend(loc='upper right')
            ax.figure.canvas.draw()
            print('Do you want to continue? (enter = yes, any other key = no)')
            answer = input().lower()
            if answer != '':
                break
        return ax

    def show(self, ax=None, cmap='nipy_spectral', fill_kw={}, onebyone=False):
        if onebyone:
            ax = self.show_onebyone(ax, cmap, fill_kw)
        else:
            ax = self.show_all(ax, cmap, fill_kw)
        return ax

    def show_slices(self, ax=None, fill_kw={}):
        ax = self._prepare_ax(ax, cmap=None)
        for spol in self.slices:
            for pol in spol.slices:
                x, y = pol.boundary.xy
                ax.fill(x, y, **fill_kw)
        return ax

    def show_polygon(self, polygon, label, ax=None, cmap='nipy_spectral', fill_kw={}):
        # print(f"Polygon:  {polygon}")
        # print(f"Label:  {label}")
        ax = self._prepare_ax(ax, cmap)
        for _ in polygon:
            x, y = zip(*polygon)
            ax.fill(x, y, label=label, **fill_kw)
        # if label not in self.added_labels:
        #     ax.legend(loc='upper right')
        #     self.added_labels.append(label)
        return ax



import setuptools
from distutils.version import StrictVersion
from importlib import import_module

with open("README.md", "r") as fh:
    long_description = fh.read()

extras = {
    'gdspy': ('gdspy', '1.6', 'pip'),
    'matplotlib': ('matplotlib', '3.2', 'conda'),
    'cycler': ('cycler', '0.10', 'conda'),
}
extras_require = {k: '>='.join(v[0:2]) for k, v in extras.items()}

install_requires = [
    'numpy>=1.18',
    'pyvista>=0.27',
]

setuptools.setup(
    name="nextnanopy",
    version="0.1.21",
    author="nextnano GmbH",
    author_email="python@nextnano.com",
    license='BSD-3-Clause',
    description="Useful tools to interface the nextnano software (https://www.nextnano.com/)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="nextnano",
    url="https://github.com/nextnanopy/nextnanopy",
    packages=setuptools.find_packages(exclude=["tests"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Intended Audience :: Customer Service",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
    ],
    python_requires='>=3.8',
    install_requires=install_requires,
    extras_require=extras_require
)

# Code below adapted from QCoDeS (https://qcodes.github.io/)

version_template = '''
*****
***** package {0} must be at least version {1}.
***** Please upgrade it (pip install -U {0} or conda install {0})
***** in order to use {2}
***** Recommended method: {3}
*****
'''

missing_template = '''
*****
***** package {0} not found
***** Please install it (pip install {0} or conda install {0})
***** Recommended: {2} install {0}
***** in order to use {1}
*****
'''

valueerror_template = '''
*****
***** package {0} version not understood
***** Please make sure the installed version ({1})
***** is compatible with the minimum required version ({2})
***** in order to use {3}
*****
'''

othererror_template = '''
*****
***** could not import package {0}. Please try importing it from
***** the commandline to diagnose the issue.
*****
'''

# now test the versions of extras
for extra, (module_name, min_version, install_method) in extras.items():
    try:
        module = import_module(module_name.lower())
        if StrictVersion(module.__version__) < StrictVersion(min_version):
            print(version_template.format(module_name, min_version, extra, install_method))
    except ImportError:
        print(missing_template.format(module_name, extra, install_method))
    except ValueError:
        print(valueerror_template.format(
            module_name, module.__version__, min_version, extra))
    except:
        print(othererror_template.format(module_name))



import unittest
from nextnanopy.utils.formatting import best_str_to_name_unit, str_to_name_unit, str_to_name_unit_with_rest


class TestNameUnit(unittest.TestCase):
    def test_str_to_name_unit(self):
        self.assertEqual(str_to_name_unit('test', init='[', end=']'), ('test', None))
        self.assertEqual(str_to_name_unit('test[', init='[', end=']'), ('test[', None))
        self.assertEqual(str_to_name_unit('test[eV', init='[', end=']'), ('test[eV', None))
        self.assertEqual(str_to_name_unit('test [eV', init='[', end=']'), ('test [eV', None))
        self.assertEqual(str_to_name_unit('test [eV]', init='[', end=']'), ('test', 'eV'))
        self.assertEqual(str_to_name_unit('test [eV] another[eV]', init='[', end=']'), ('test', 'eV'))
        self.assertEqual(str_to_name_unit('test eV] ', init='[', end=']'), ('test eV]', None))
        self.assertEqual(str_to_name_unit('test[eV]_weird', init='[', end=']', add_rest_to_name=False), ('test', 'eV'))
        self.assertEqual(str_to_name_unit('test[eV]_weird', init='[', end=']', add_rest_to_name=True),
                         ('test_weird', 'eV'))
        self.assertEqual(str_to_name_unit('test [eV] _weird', init='[', end=']', add_rest_to_name=False),
                         ('test', 'eV'))
        self.assertEqual(str_to_name_unit('test [eV] _weird', init='[', end=']', add_rest_to_name=True),
                         ('test _weird', 'eV'))

        self.assertEqual(str_to_name_unit('test (eV]', init='[', end=']'), ('test (eV]', None))
        self.assertEqual(str_to_name_unit('test (eV)', init='[', end=']'), ('test (eV)', None))
        self.assertEqual(str_to_name_unit('test (eV)', init='(', end=')'), ('test', 'eV'))

    def test_str_to_name_unit_with_rest(self):
        self.assertEqual(str_to_name_unit_with_rest('test', init='[', end=']'), ('test', None))
        self.assertEqual(str_to_name_unit_with_rest('test[', init='[', end=']'), ('test[', None))
        self.assertEqual(str_to_name_unit_with_rest('test[eV', init='[', end=']'), ('test[eV', None))
        self.assertEqual(str_to_name_unit_with_rest('test [eV', init='[', end=']'), ('test [eV', None))
        self.assertEqual(str_to_name_unit_with_rest('test [eV]', init='[', end=']'), ('test', 'eV'))
        self.assertEqual(str_to_name_unit_with_rest('test [eV] another[eV]', init='[', end=']'), ('test another', 'eV'))
        self.assertEqual(str_to_name_unit_with_rest('test eV] ', init='[', end=']'), ('test eV]', None))
        self.assertEqual(str_to_name_unit_with_rest('test[eV]_weird', init='[', end=']'), ('test_weird', 'eV'))
        self.assertEqual(str_to_name_unit_with_rest('test [eV] _weird', init='[', end=']'), ('test _weird', 'eV'))

        self.assertEqual(str_to_name_unit_with_rest('test (eV]', init='[', end=']'), ('test (eV]', None))
        self.assertEqual(str_to_name_unit_with_rest('test (eV)', init='[', end=']'), ('test (eV)', None))
        self.assertEqual(str_to_name_unit_with_rest('test (eV)', init='(', end=')'), ('test', 'eV'))

    def test_best_str_to_name_unit(self):
        self.assertEqual(best_str_to_name_unit('test'), ('test', None))
        self.assertEqual(best_str_to_name_unit('test['), ('test[', None))
        self.assertEqual(best_str_to_name_unit('test[eV'), ('test[eV', None))
        self.assertEqual(best_str_to_name_unit('test [eV'), ('test [eV', None))
        self.assertEqual(best_str_to_name_unit('test [eV]'), ('test', 'eV'))
        self.assertEqual(best_str_to_name_unit('test [eV] another[eV]'), ('test another', 'eV'))
        self.assertEqual(best_str_to_name_unit('test eV] '), ('test eV]', None))
        self.assertEqual(best_str_to_name_unit('test[eV]_weird'), ('test_weird', 'eV'))
        self.assertEqual(best_str_to_name_unit('test [eV] _weird'), ('test _weird', 'eV'))

        self.assertEqual(best_str_to_name_unit('test (eV]'), ('test (eV]', None))
        self.assertEqual(best_str_to_name_unit('test (eV)'), ('test', 'eV'))
        self.assertEqual(best_str_to_name_unit('test (eV)'), ('test', 'eV'))


if __name__ == '__main__':
    unittest.main()



from nextnanopy.inputs import InputFileTemplate
from nextnanopy.msb.defaults import is_msb_input_text


class InputFile(InputFileTemplate):
    def load_variables(self):
        pass

    def validate(self):
        if not is_msb_input_text(self.raw_text):
            raise ValueError(f'Not valid nextnano.MSB input file')
